//----------------------------------------------------------------------
// Includes
//----------------------------------------------------------------------
#include "MantidDataHandling/LoadCSNSNexus.h"
#include "MantidDataHandling/EventWorkspaceCollection.h"

#include <boost/random/mersenne_twister.hpp>
#include <boost/random/uniform_real.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/shared_array.hpp>
#include <boost/scoped_array.hpp>
#include <boost/function.hpp>
#include <functional>

#include "MantidKernel/ArrayProperty.h"
#include "MantidKernel/ThreadPool.h"
#include "MantidKernel/UnitFactory.h"
#include "MantidKernel/ThreadSchedulerMutexes.h"
#include "MantidKernel/BoundedValidator.h"
#include "MantidKernel/VisibleWhenProperty.h"
#include "MantidKernel/TimeSeriesProperty.h"
#include "MantidGeometry/Instrument/RectangularDetector.h"
#include "MantidAPI/FileProperty.h"
#include "MantidAPI/MemoryManager.h"
#include "MantidAPI/RegisterFileLoader.h"
#include "MantidAPI/SpectrumDetectorMapping.h"
#include "MantidKernel/Timer.h"

using std::cout;
using std::endl;
using std::map;
using std::string;
using std::vector;

namespace Mantid
{
namespace DataHandling
{

DECLARE_NEXUS_FILELOADER_ALGORITHM(LoadCSNSNexus)

using namespace Kernel;
using namespace Geometry;
using namespace API;
using namespace DataObjects;

LoadCSNSNexus::LoadCSNSNexus()
    : m_bNumPixels(0),m_mNumPixels(0), pulseTimes(0), m_numBins(0), m_spec_min(0),
      m_spec_max(0), m_fileMutex() {}

//-------------------------------------------------------------------------------------------------
/// Initialisation method.
void LoadCSNSNexus::init()
{
    declareProperty(
        new FileProperty("Filename", "", FileProperty::Load, {".nxs"}),
        "The name of the NeXus file to load");
    declareProperty(new WorkspaceProperty<API::Workspace>("OutputWorkspace", "",
                    Direction::Output),
                    "The name of the Workspace2D to create.");
    declareProperty(
        new PropertyWithValue<string>("Entryname", "csns",Direction::Input), "Optional: Name of entry (default csns)");

    declareProperty(
        new ArrayProperty<string>("Bankname", Direction::Input),
        "Optional: A comma-separated list of bank to read. (default all banks)");
    declareProperty(
        new ArrayProperty<string>("Monitorname", Direction::Input),
        "Optional: A comma-separated list of monitor to read. (default all monitors)");
    declareProperty(
        new PropertyWithValue<bool>("Loadevent", true, Direction::Input),
        "Default true: load event data mode, false: Load histogram data.");

    auto mustBePositive = boost::make_shared<BoundedValidator<int>>();
    mustBePositive->setLower(1);
    declareProperty(
        new PropertyWithValue<specid_t>("SpectrumMin", 1, mustBePositive),
        "The index number of the first spectrum to read.  Only used if spectrum_max is set.");
    declareProperty(
        new PropertyWithValue<specid_t>("SpectrumMax", Mantid::EMPTY_INT(),
                                        mustBePositive),
        "The number of the last spectrum to read. Only used if explicitly set.");
}

//-------------------------------------------------------------------------------------------------
/**
 * Return the confidence with with this algorithm can load the file
 * @param descriptor A descriptor for the file
 * @returns An integer specifying the confidence level. 0 indicates it will not
 * be used
 */
int LoadCSNSNexus::confidence(Kernel::NexusDescriptor &descriptor) const
{
    int confidence(0);
    if (descriptor.pathOfTypeExists("/csns", "NXentry"))
    {
        const bool hasEventData = descriptor.classTypeExists("NXevent_data");
        const bool hasHistogramData = descriptor.classTypeExists("NXdata");
        if (hasHistogramData && hasEventData)
            // Event data = this is event NXS
            confidence = 20;
        else if (hasHistogramData && !hasEventData)
            // No event data = this is the one
            confidence = 80;
        else if (!hasHistogramData && hasEventData)
            // Histogram Data only
            confidence = 40;
        else
            // No data ?
            confidence = 10;
    }
    return confidence;
}

//-------------------------------------------------------------------------------------------------
/**
 * Return entry name
 * @param[in] file :: nxs file
 * @returns newEntry :: entry name
 */
std::string LoadCSNSNexus::getEntryName(::NeXus::File *file)
{
    std::string newEntry="";

    std::map<std::string, std::string> entries = file->getEntries();

    if (entries.empty())
    {
        throw std::runtime_error("No entries in the NXS file!");
    }
    else if (entries.find("csns") != entries.end())
    {
        newEntry = "csns";
    }
    else
    {
        newEntry = entries.begin()->first;
    }

    return newEntry;
}

//-------------------------------------------------------------------------------------------------
/**
* Return the pixel number
* @param[in] file :: nxs file
* @return numPixels
*/
size_t LoadCSNSNexus::getNumPixels(::NeXus::File *file, bool bm)
{
    size_t numPixels=0;
    size_t newPixels=0;
    size_t numBank = m_bankNames.size();
    size_t numMonitor = m_monitorNames.size();

    std::map<std::string, std::string> entries;
    std::map<std::string, std::string> subentries;
    std::map<std::string, std::string>::iterator it;
    
    file->openGroup(m_entryName, "NXentry");
    file->openGroup("instrument", "NXinstrument");
    if (bm)
    {
        if (numBank==0)
        {
            // Look for all the banks
            entries = file->getEntries();
            for (it = entries.begin(); it != entries.end(); ++it)
            {
                std::string name = it->first;

                if (name.substr(0, 4) == "bank")
                {
                    // cout<<"name is:"<<name<<endl;
                    // file->openGroup(name, "NXdetector");
                    file->openGroup(name, it->second);

                    subentries = file->getEntries();

                    if (subentries.find("pixel_id") != subentries.end())
                    {
                        file->openData("pixel_id");
                        std::vector<int64_t> dims = file->getInfo().dims;
                        file->closeData();

                        if (!dims.empty())
                        {
                            newPixels = 1;
                            for (size_t i = 0; i < dims.size(); i++)
                                newPixels *= dims[i];
                            // cout<<"newPixels is: "<<newPixels<<endl;
                            numPixels += newPixels;
                        }
                        // cout<<"numPixels is:"<<numPixels<<endl;
                    }
                    else
                    {
                        throw std::runtime_error("No pixels in the NXS file!");
                    }
                    file->closeGroup();
                }
            }
        }else{
            // Look for all the banks
            for (auto j = m_bankNames.begin(); j != m_bankNames.end(); ++j)
            {
                std::string name = *j;
                file->openGroup(name, "NXdetector");
                
                entries = file->getEntries();
                if (entries.find("pixel_id") != entries.end())
                {
                    file->openData("pixel_id");
                    std::vector<int64_t> dims = file->getInfo().dims;
                    file->closeData();
                    if (!dims.empty())
                    {
                        newPixels = 1;
                        for (size_t i = 0; i < dims.size(); i++)
                            newPixels *= dims[i];
                        numPixels += newPixels;
                    }
                }
                else
                {
                    throw std::runtime_error("No pixels in the banks file!");
                }
                file->closeGroup();
            }
        }
    }
    else
    {
        if (numMonitor==0)
        {
            // Look for all the banks
            entries = file->getEntries();
            for (it = entries.begin(); it != entries.end(); ++it)
            {
                std::string name = it->first;

                if (name.substr(0, 7) == "monitor")
                {
                    // cout<<"name is:"<<name<<endl;
                    // file->openGroup(name, "NXdetector");
                    file->openGroup(name, it->second);

                    subentries = file->getEntries();

                    if (subentries.find("pixel_id") != subentries.end())
                    {
                        file->openData("pixel_id");
                        std::vector<int64_t> dims = file->getInfo().dims;
                        file->closeData();

                        if (!dims.empty())
                        {
                            newPixels = 1;
                            for (size_t i = 0; i < dims.size(); i++)
                                newPixels *= dims[i];
                            // cout<<"newPixels is: "<<newPixels<<endl;
                            numPixels += newPixels;
                        }
                        // cout<<"numPixels is:"<<numPixels<<endl;
                    }
                    else
                    {
                        throw std::runtime_error("No pixels in the NXS file!");
                    }
                    file->closeGroup();
                }
            }
        }else{
            // Look for all the monitors
            for (auto j = m_monitorNames.begin(); j != m_monitorNames.end(); ++j)
            {
                std::string name = *j;
                file->openGroup(name, "NXdetector");
                entries = file->getEntries();

                if (entries.find("pixel_id") != entries.end())
                {
                    file->openData("pixel_id");
                    std::vector<int64_t> dims = file->getInfo().dims;
                    file->closeData();
                    if (!dims.empty())
                    {
                        newPixels = 1;
                        for (size_t i = 0; i < dims.size(); i++)
                            newPixels *= dims[i];
                        numPixels += newPixels;
                    }
                }
                else
                {
                    throw std::runtime_error("No pixels in the monitors file!");
                }
                file->closeGroup();
            }
        }
    }

    file->closeGroup();
    file->closeGroup();
    return numPixels;
}

//-------------------------------------------------------------------------------------------------
/**
* Return the pixel number
* @param[in] filen :: nxs file
* @return numBins :: bins of TOF
*/
size_t LoadCSNSNexus::getNumBins(::NeXus::File *file, bool bm)
{
    size_t numBins=0;
    size_t numBank = m_bankNames.size();
    size_t numMonitor = m_monitorNames.size();

    std::map<std::string, std::string> entries;
    std::map<std::string, std::string> subentries;
    std::map<std::string, std::string>::iterator it;

    file->openGroup(m_entryName, "NXentry");
    file->openGroup("instrument", "NXinstrument");
    if (bm)
    {
        if (numBank==0)
        {
            // Look for all the banks
            entries = file->getEntries();
            for (it = entries.begin(); it != entries.end(); ++it)
            {
                std::string name = it->first;

                if (name.substr(0, 4) == "bank")
                {
                    file->openGroup(name, it->second);
                    subentries = file->getEntries();

                    if (subentries.find("time_of_flight") != subentries.end())
                    {
                        file->openData("time_of_flight");
                        std::vector<int64_t> dims = file->getInfo().dims;
                        file->closeData();
                        numBins = dims[0];
                        break;
                    }else{
                        throw std::runtime_error("No pixels in the NXS file!");
                    }
                    file->closeGroup();
                    break;
                }
             }
        }else{
             for (auto j = m_bankNames.begin(); j != m_bankNames.end(); ++j)
             {
                std::string name = *j;
                file->openGroup(name, "NXdetector");

                entries = file->getEntries();
                if (entries.find("time_of_flight") != entries.end())
                {
                    file->openData("time_of_flight");
                    std::vector<int64_t> dims = file->getInfo().dims;
                    file->closeData();
                    numBins = dims[0];
                }else{
                    throw std::runtime_error("No pixels in the NXS file!");
                }
                file->closeGroup();
                break;
              }
            }
    }else{
         if (numMonitor==0)
        {
            // Look for all the banks
            entries = file->getEntries();
            for (it = entries.begin(); it != entries.end(); ++it)
            {
                std::string name = it->first;

                if (name.substr(0, 7) == "monitor")
                {
                    file->openGroup(name, it->second);
                    subentries = file->getEntries();

                    if (subentries.find("time_of_flight") != subentries.end())
                    {
                        file->openData("time_of_flight");
                        std::vector<int64_t> dims = file->getInfo().dims;
                        file->closeData();
                        numBins = dims[0];
                    }else{
                        throw std::runtime_error("No pixels in the NXS file!");
                    }
                file->closeGroup();
                break;
                }
             }
        }else{
             for (auto j = m_bankNames.begin(); j != m_bankNames.end(); ++j)
             {
                std::string name = *j;
                file->openGroup(name, "NXdetector");

                entries = file->getEntries();
                if (entries.find("time_of_flight") != entries.end())
                {
                    file->openData("time_of_flight");
                    std::vector<int64_t> dims = file->getInfo().dims;
                    file->closeData();
                    numBins = dims[0];
                }else{
                    throw std::runtime_error("No pixels in the NXS file!");
                }
                file->closeGroup();
                break;
                }
            }
}
    file->closeGroup();
    file->closeGroup();
    file->closeGroup();
    return numBins;
}
//-------------------------------------------------------------------------------------------------
/**
* Return the event data
* @param[in] file :: nxs file
* @param[in] bm :: true is bank and false is monitor
* @return eventData :: event data include event_pixel_id, pulse_time,tof.
*/
std::vector<CSNSEventData> LoadCSNSNexus::getEventData(::NeXus::File *file, bool bm)
{
    size_t numBank = m_bankNames.size();
    size_t numMonitor = m_monitorNames.size();

    std::map<std::string, std::string> entries;
    std::map<std::string, std::string> subentries;
    std::map<std::string, std::string>::iterator it;
    std::vector<int> _pixelId;
    std::vector<float> _timeOfFlight;
    std::vector<float> _pulseTime;
    CSNSEventData _eventDatatmp;    
    std::vector<CSNSEventData> eventData;

    file->openGroup(m_entryName, "NXentry");
    file->openGroup("event_data", "NXcollection");
    if (bm)
    {
        if (numBank==0)
        {
            entries = file->getEntries();
            for (it = entries.begin(); it != entries.end(); ++it)
            {
                std::string name = it->first;

                if (name.substr(0, 4) == "bank")
                {
                    file->openGroup(name, it->second);
                    file->readData("event_pixel_id",_pixelId);
                    file->readData("event_time_of_flight",_timeOfFlight);
                    file->readData("event_pulse_time",_pulseTime);

                    for (size_t i=0; i<_pixelId.size(); ++i)
                    {
                        _eventDatatmp.pixelId = _pixelId[i];
                        _eventDatatmp.pulseTime = _pulseTime[i];
                        _eventDatatmp.tof = _timeOfFlight[i];
                        eventData.push_back(_eventDatatmp);
                    }
                    file->closeGroup();
                }
             }
        }else{
             for (auto j = m_bankNames.begin(); j != m_bankNames.end(); ++j)
             {
                std::string name = *j;
                file->openGroup(name, "NXevent_data");
                    
                file->readData("event_pixel_id",_pixelId);
                file->readData("event_time_of_flight",_timeOfFlight);
                file->readData("event_pulse_time",_pulseTime);
                   
                for (size_t i=0; i<_pixelId.size(); ++i)
                {
                    _eventDatatmp.pixelId = _pixelId[i];
                    _eventDatatmp.pulseTime = _pulseTime[i];
                    _eventDatatmp.tof = _timeOfFlight[i];
                    eventData.push_back(_eventDatatmp);
                }
                    file->closeGroup();

              }
            }
    }else{
         if (numMonitor==0)
         {
            entries = file->getEntries();
            for (it = entries.begin(); it != entries.end(); ++it)
            {
                std::string name = it->first;

                if (name.substr(0, 7) == "monitor")
                {
                    file->openGroup(name, it->second);
                    file->readData("event_pixel_id",_pixelId);
                    file->readData("event_time_of_flight",_timeOfFlight);
                    file->readData("event_pulse_time",_pulseTime);

                    for (size_t i=0; i<_pixelId.size(); ++i)
                    {
                        _eventDatatmp.pixelId = _pixelId[i];
                        _eventDatatmp.pulseTime = _pulseTime[i];
                        _eventDatatmp.tof = _timeOfFlight[i];
                        eventData.push_back(_eventDatatmp);
                    }
                    file->closeGroup();
                }
             }
        }else{
             for (auto j = m_monitorNames.begin(); j != m_monitorNames.end(); ++j)
             {
                std::string name = *j;
                file->openGroup(name, "NXevent_data");

                file->readData("event_pixel_id",_pixelId);
                file->readData("event_time_of_flight",_timeOfFlight);
                file->readData("event_pulse_time",_pulseTime);

                for (size_t i=0; i<_pixelId.size(); ++i)
                {
                    _eventDatatmp.pixelId = _pixelId[i];
                    _eventDatatmp.pulseTime = _pulseTime[i];
                    _eventDatatmp.tof = _timeOfFlight[i];
                    eventData.push_back(_eventDatatmp);
                }
                    file->closeGroup();

              }
              }
         }
    file->closeGroup();
    file->closeGroup();
    return eventData;
}
//-------------------------------------------------------------------------------------------------
/**
* Return eventPixelId vector
* @param[in] file :: nxs file
* @return eventPixelId :: Event pixel id list

std::vector<int> LoadCSNSNexus::getEventPixelId(::NeXus::File *file, bool bm)
{
    std::vector<int> _tmp;;
    std::vector<int> eventPixelId;;
    size_t numBank = m_bankNames.size();
    size_t numMonitor = m_monitorNames.size();

    file->openGroup(m_entryName,"NXentry");
    file->openGroup("event_data","NXcollection");
    std::map<std::string, std::string> entries;
    std::map<std::string, std::string>::iterator it;

    if (bm)
    {
        if (numBank==0)
        {
            // Look for all the banks
            entries = file->getEntries();
            for (it = entries.begin(); it != entries.end(); ++it)
            {
                std::string name = it->first;
                if (name.substr(0, 4) == "bank")
                {
                    file->openGroup(name, it->second);
                    file->readData("event_pixel_id",_tmp);
                    int pid;
                    for (size_t k = 0; k<_tmp.size();++k)
                    {
                        pid=_tmp[k];
                        eventPixelId.push_back(pid);
                    }
                    file->closeGroup();
                }
             }
        }else{
                for (auto j = m_bankNames.begin(); j != m_bankNames.end(); ++j)
                {
                    std::string name = *j;
                    file->openGroup(name, "NXevent_data");
                    file->readData("event_pixel_id",_tmp);
                    int pid;
                    for (size_t k = 0; k<_tmp.size();++k)
                    {
                        pid=_tmp[k];
                        eventPixelId.push_back(pid);
                    }
                    file->closeGroup();
                }
            }

    }else{
            if (numMonitor==0)
            {
                // Look for all the monitor
                entries = file->getEntries();
                for (it = entries.begin(); it != entries.end(); ++it)
                {
                    std::string name = it->first;
                    if (name.substr(0, 7) == "monitor")
                    {
                        file->openGroup(name, it->second);
                        file->readData("event_pixel_id",_tmp);
                        int pid;
                        for (size_t k = 0; k<_tmp.size();++k)
                        {
                            pid=_tmp[k];
                            eventPixelId.push_back(pid);
                        }
                        file->closeGroup();
                    }
                }
             }else{
                    for (auto j = m_monitorNames.begin(); j != m_monitorNames.end(); ++j)
                    {
                        std::string name = *j;
                        file->openGroup(name, "NXevent_data");
                        file->readData("event_pixel_id",_tmp);
                        int pid;
                        for (size_t k = 0; k<_tmp.size();++k)
                        {
                            pid=_tmp[k];
                            eventPixelId.push_back(pid);
                        }
                        file->closeGroup();
                    }
                }
        }
        file->closeGroup();
        file->closeGroup();
        return eventPixelId;
}
*/

//-------------------------------------------------------------------------------------------------
/**
* Return Bank list NX/entry/instrument/bank*
* @param[in] file :: nxs file
* @return bankNames :: bank list
*/
std::vector<std::string> LoadCSNSNexus::getBankNames(::NeXus::File *file)
{
    std::vector<std::string> bankNames;

    // Open the default data group 'entry'
    file->openGroup(m_entryName, "NXentry");
    // Also pop into the instrument
    file->openGroup("instrument", "NXinstrument");

    // Look for all the banks
    std::map<std::string, std::string> entries = file->getEntries();
    std::map<std::string, std::string>::iterator it;
    for (it = entries.begin(); it != entries.end(); ++it)
    {
        if (it->first.substr(0,4)=="bank" && it->second=="NXdetector")
        {
            bankNames.push_back(it->first);
        }
    }

    file->closeGroup();
    file->closeGroup();
    return bankNames;
}
//-------------------------------------------------------------------------------------------------
/**
* Return the pixel offset
* @param[in] file :: nxs file
* @return pixelOffset :: [x,y]
*/

std::vector<std::vector<float> > LoadCSNSNexus::getPixelOffset(::NeXus::File *file, const std::string &bankName)
{
    std::vector<std::vector<float> > pixelOffset;

    file->openGroup(m_entryName, "NXentry");
    file->openGroup("instrument", "NXinstrument");
    file->openGroup(bankName, "NXdetector");

    std::vector<float> _tmp;
    file->readData("x_pixel_offset",_tmp);
    pixelOffset.push_back(_tmp);
    file->readData("y_pixel_offset",_tmp);
    pixelOffset.push_back(_tmp);

    file->closeGroup();
    file->closeGroup();
    file->closeGroup();

    return pixelOffset;
}
//-------------------------------------------------------------------------------------------------
/**
 * Creates a time series property from the currently opened log entry. It is
 * assumed to
 * have been checked to have a time field and the value entry's name is given as
 * an argument
 * @param file :: A reference to the file handle
 * @param prop_name :: The name of the property
 * @returns A pointer to a new property containing the time series
 */
Kernel::Property *LoadCSNSNexus::createTimeSeries(::NeXus::File *file,
        const std::string &prop_name) const
{
    file->openData("time");
    //----- Start time is an ISO8601 string date and time. ------
    std::string start;
    try
    {
        file->getAttr("start", start);
    }
    catch (::NeXus::Exception &)
    {
        start="2017-12-31T00:00:00+08:00";
    }

    // Convert to date and time
    Kernel::DateAndTime start_time = Kernel::DateAndTime(start);
    std::string time_units;
    file->getAttr("units", time_units);
    if (time_units.compare("second") < 0 && time_units != "s" &&
            time_units != "minutes") // Can be s/second/seconds/minutes
    {
        file->closeData();
        throw ::NeXus::Exception("Unsupported time unit '" + time_units + "'");
    }
    //--- Load the seconds into a double array ---
    std::vector<double> time_double;
    try
    {
        file->getDataCoerce(time_double);
    }
    catch (::NeXus::Exception &e)
    {
        g_log.warning() << "Log entry's time field could not be loaded: '"
                        << e.what() << "'.\n";
        file->closeData();
        throw;
    }
    file->closeData(); // Close time data
    g_log.debug() << "   done reading \"time\" array\n";
    // Convert to seconds if needed
    if (time_units == "minutes")
    {
        std::transform(time_double.begin(), time_double.end(), time_double.begin(),
                       std::bind2nd(std::multiplies<double>(), 60.0));
    }
    // Now the values: Could be a string, int or double
    file->openData("value");
    // Get the units of the property
    std::string value_units("");
    try
    {
        file->getAttr("units", value_units);
    }
    catch (::NeXus::Exception &)
    {
        // Ignore missing units field.
        value_units = "";
    }

    // Now the actual data
    ::NeXus::Info info = file->getInfo();
    // Check the size
    if (size_t(info.dims[0]) != time_double.size())
    {
        file->closeData();
        throw ::NeXus::Exception("Invalid value entry for time series");
    }
    if (file->isDataInt()) // Int type
    {
        std::vector<int> values;
        try
        {
            file->getDataCoerce(values);
            file->closeData();
        }
        catch (::NeXus::Exception &)
        {
            file->closeData();
            throw;
        }
        // Make an int TSP
        auto tsp = new TimeSeriesProperty<int>(prop_name);
        tsp->create(start_time, time_double, values);
        tsp->setUnits(value_units);
        g_log.debug() << "   done reading \"value\" array\n";
        return tsp;
    }
    else if (info.type == ::NeXus::CHAR)
    {
        std::string values;
        const int64_t item_length = info.dims[1];
        try
        {
            const int64_t nitems = info.dims[0];
            const int64_t total_length = nitems * item_length;
            boost::scoped_array<char> val_array(new char[total_length]);
            file->getData(val_array.get());
            file->closeData();
            values = std::string(val_array.get(), total_length);
        }
        catch (::NeXus::Exception &)
        {
            file->closeData();
            throw;
        }
        // The string may contain non-printable (i.e. control) characters, replace
        // these
        std::replace_if(values.begin(), values.end(), iscntrl, ' ');
        auto tsp = new TimeSeriesProperty<std::string>(prop_name);
        std::vector<DateAndTime> times;
        DateAndTime::createVector(start_time, time_double, times);
        const size_t ntimes = times.size();
        for (size_t i = 0; i < ntimes; ++i)
        {
            std::string value_i =
                std::string(values.data() + i * item_length, item_length);
            tsp->addValue(times[i], value_i);
        }
        tsp->setUnits(value_units);
        g_log.debug() << "   done reading \"value\" array\n";
        return tsp;
    }
    else if (info.type == ::NeXus::FLOAT32 || info.type == ::NeXus::FLOAT64)
    {
        std::vector<double> values;
        try
        {
            file->getDataCoerce(values);
            file->closeData();
        }
        catch (::NeXus::Exception &)
        {
            file->closeData();
            throw;
        }
        auto tsp = new TimeSeriesProperty<double>(prop_name);
        tsp->create(start_time, time_double, values);
        tsp->setUnits(value_units);
        g_log.debug() << "   done reading \"value\" array\n";
        return tsp;
    }
    else
    {
        throw ::NeXus::Exception(
            "Invalid value type for time series. Only int, double or strings are "
            "supported");
    }
}
/**
* Return the pixel_id
* @param[in] file :: nxs file
* @param[in] bankName :: bank name
* @return pixelId :: pixel_id
*/

std::vector<uint32_t> LoadCSNSNexus::getPixelId(::NeXus::File *file, const std::string &bankName)
{
    std::vector<int> tmp;
    file->openGroup(m_entryName, "NXentry");
    file->openGroup("instrument", "NXinstrument");
    file->openGroup(bankName, "NXdetector");

    file->readData("pixel_id",tmp);
    std::vector<uint32_t> pixelId(tmp.begin(), tmp.end());

    file->closeGroup();
    file->closeGroup();
    file->closeGroup();

    return pixelId;
}

//-------------------------------------------------------------------------------------------------
/**
 * Load an log entry
 * @param file:: NeXus file
 * @param workspace :: A pointer to the workspace to store the logs
 */
void LoadCSNSNexus::loadNXLog(
    ::NeXus::File *file,
    boost::shared_ptr<API::MatrixWorkspace> workspace) const
{

    file->openGroup(m_entryName, "NXentry");

    try
    {
        file->openGroup("logs", "NXcollection");
    }
    catch (::NeXus::Exception &e)
    {
        g_log.warning() <<"Read logs: " << e.what()<<"\n";
        file->closeGroup();
        return;
    }

    // Validate the NX log class.
    std::map<std::string, std::string> entries = file->getEntries();
    std::map<std::string, std::string>::iterator it;
    for (it = entries.begin(); it != entries.end(); ++it)
    {
        std::string logName=it->first;
        if (it->second == "NXlog")
        {
            file->openGroup(it->first, "NXlog");

            std::map<std::string, std::string> subEntries = file->getEntries();
            if ((subEntries.find("value") == subEntries.end()) || (subEntries.find("time") == subEntries.end()))
            {
                g_log.warning() << "Invalid NXlog entry " << logName
                                << " found. Did not contain 'value' and 'time'.\n";
            }
            else
            {
                try
                {
                    if (!(workspace->run().hasProperty(logName)))
                    {
                        g_log.debug()<<logName<<"\n";
                        Kernel::Property *logValue = createTimeSeries(file, logName);
                        workspace->mutableRun().addProperty(logValue, true);
                    }
                }
                catch (::NeXus::Exception &e)
                {
                    g_log.warning() << "NXlog entry " << logName
                                    << " gave an error when loading:'" << e.what() << "'.\n";
                }

            }
            file->closeGroup();
        }//if
        if (it->second == "NXcollection")
        {
            std::string groupName=it->first;
            file->openGroup(it->first, "NXcollection");

            std::map<std::string, std::string> subEntries = file->getEntries();
            std::map<std::string, std::string>::iterator subit;
            for (subit = subEntries.begin(); subit != subEntries.end(); ++subit)
            {
                if (subit->second == "NXlog")
                {
                    file->openGroup(subit->first, "NXlog");
                    std::map<std::string, std::string> subEntries = file->getEntries();
                    if ((subEntries.find("value") == subEntries.end()) || (subEntries.find("time") == subEntries.end()))
                    {
                        g_log.warning() << "Invalid NXlog entry " << subit->first
                                        << " found. Did not contain 'value' and 'time'.\n";
                    }
                    else
                    {
                        try
                        {
                            if (!(workspace->run().hasProperty(subit->first)))
                            {
                                g_log.debug()<<subit->first<<"\n";
                                Kernel::Property *logValue = createTimeSeries(file, groupName+"_"+subit->first);
                                workspace->mutableRun().addProperty(logValue, true);
                            }
                        }
                        catch (::NeXus::Exception &e)
                        {
                            g_log.warning() << "NXlog entry " << subit->first
                                            << " gave an error when loading:'" << e.what() << "'.\n";
                        }

                    }
                    file->closeGroup();
                }

            }
            file->closeGroup();
        }//if
    }//for
    file->closeGroup();
    file->closeGroup();
    std::cout<<"current path: "<<file->getPath()<<std::endl;
}

//-------------------------------------------------------------------------------------------------
/**
* Return the CSNS NeXus definiton version
* @param[in] filename :: nxs file name
* @return version :: CSNS NeXus definition version
*/
std::string LoadCSNSNexus::getVersion(::NeXus::File *file)
{
    std::string version="";

    file->openGroup(m_entryName, "NXentry");

    std::map<std::string, std::string> entries = file->getEntries();

    if (entries.find("version") != entries.end())
    {
        file->readData("version", version);
    }
    else
    {
        throw std::runtime_error("No version in the NXS file!");
    }

    file->closeGroup();
    return version;
}

//-------------------------------------------------------------------------------------------------
/**
* Return true if has event data
* @param[in] file :: nxs file
* @return hasEventData :: true|false
*/
bool LoadCSNSNexus::hasEventData(::NeXus::File *file)
{
    bool hasEventData=false;

    file->openGroup(m_entryName, "NXentry");
    try
    {
        file->openGroup("event_data", "NXcollection");
        std::map<std::string, std::string> entries = file->getEntries();
        std::map<std::string, std::string>::iterator it;
        for (it = entries.begin(); it != entries.end(); ++it)
        {
            if (it->second=="NXevent_data")
            {
                hasEventData = true;
                break;
            }
        }
        file->closeGroup();

    }
    catch(::NeXus::Exception)
    {
        hasEventData=false;
    }

    file->closeGroup();
    return hasEventData;
}

//-------------------------------------------------------------------------------------------------
/**
* Return true if has histogram data
* @param[in] file :: nxs file
* @return hasEventData :: true|false
*/
bool LoadCSNSNexus::hasHistogramData(::NeXus::File *file)
{
    bool hasHistogramData=false;

    file->openGroup(m_entryName, "NXentry");
    try
    {
        file->openGroup("histogram_data", "NXcollection");
        std::map<std::string, std::string> entries = file->getEntries();
        std::map<std::string, std::string>::iterator it;
        for (it = entries.begin(); it != entries.end(); ++it)
        {
            if (it->second=="NXdata")
            {
                hasHistogramData = true;
                break;
            }
        }
        file->closeGroup();

    }
    catch(::NeXus::Exception)
    {
        hasHistogramData=false;
    }

    file->closeGroup();
    return hasHistogramData;
}

//-------------------------------------------------------------------------------------------------
/**
* Return the start time of experiment
* @param[in] filename :: nxs file name
* @return startTime :: start time
*/
Kernel::DateAndTime LoadCSNSNexus::getStartTime(::NeXus::File *file)
{
    std::string startTime;

    file->openGroup(m_entryName, "NXentry");

    std::map<std::string, std::string> entries = file->getEntries();

    if (entries.find("start_time") != entries.end())
    {
        file->readData("start_time", startTime);
    }
    else
    {
        throw std::runtime_error("No start time in the NXS file!");
    }
    file->closeGroup();
    return Kernel::DateAndTime(startTime);;
}

//-------------------------------------------------------------------------------------------------
/**
* Return the end time of experiment
* @param[in] filename :: nxs file name
* @return endTime :: end time
*/
Kernel::DateAndTime LoadCSNSNexus::getEndTime(::NeXus::File *file)
{
    std::string endTime;

    file->openGroup(m_entryName, "NXentry");

    std::map<std::string, std::string> entries = file->getEntries();

    if (entries.find("end_time") != entries.end())
    {
        file->readData("end_time", endTime);
    }
    else
    {
        throw std::runtime_error("No end time in the NXS file!");
    }
    file->closeGroup();
    return Kernel::DateAndTime(endTime);
}

void LoadCSNSNexus::exec()
{
    // The input properties
    std::string filename = getPropertyValue("Filename");
    m_entryName = getPropertyValue("Entryname");
    m_bankNames = getProperty("Bankname");
    m_monitorNames = getProperty("Monitorname");
    /*
        for (auto i = m_bankNames.begin(); i != m_bankNames.end(); ++i)
        {
            std::cout<<"bank names are: "<<*i<<std::endl;
        }
    */
    bool m_loadEvent = getProperty("Loadevent");
    if (m_loadEvent)
    {
        g_log.debug() <<"Event data will be loaded.\n";
    }
    else
    {
        g_log.debug() <<"Histogram data will be loaded.\n";
    }
    m_spec_min = getProperty("SpectrumMin");
    m_spec_max = getProperty("SpectrumMax");

    auto prog = new Progress(this, 0.0, 1.0, 10);
    prog->doReport("Start to load Nexus");

    // Open nxs file
    auto file = new ::NeXus::File(filename);

    // Find the entry name we want.
    std::string newEntryName = getEntryName(file);
    if (newEntryName != m_entryName)
    {
        throw std::runtime_error(m_entryName+" not found in the NXS file!");
    }

    // Get CSNS Nexus Version
    std::string m_version = getVersion(file);
    g_log.debug()<<"CSNS Nexus version is: " <<m_version<<"\n";

    // Get start time
    Kernel::DateAndTime start_time = getStartTime(file);
    cout<<"start_time is: " <<start_time<<"\n";
    
    // Get end time
    Kernel::DateAndTime end_time = getEndTime(file);
    cout<<"end_time is: " <<end_time<<"\n";

    // has Event data
    if (hasEventData(file))
    {
        cout<<"has Event data\n";
    }

    // has Histogram data
    if (hasHistogramData(file))
    {
        cout<<"has Histogram data\n";
    }

    prog->doReport("Counting pixels");
    m_bNumPixels = getNumPixels(file, true);
    m_mNumPixels = getNumPixels(file, false);
    g_log.debug() <<m_bNumPixels <<" pixels found."<< std::endl;
    std::cout<<"num of pix in banks are: "<<m_bNumPixels<<std::endl;
    std::cout<<"num of pix in monitors are: "<<m_mNumPixels<<std::endl;

    prog->doReport("Counting TOF bins");
    m_bNumBins = getNumBins(file, true);
    m_mNumBins = getNumBins(file, false);
    g_log.debug() <<m_numBins <<" TOF bins found."<< std::endl;
    std::cout<<"num of bin in banks are: "<<m_bNumBins<<std::endl;
    std::cout<<"num of bin in monitors are: "<<m_mNumBins<<std::endl;

    std::vector<std::string> bankNames= getBankNames(file);
    g_log.debug() <<bankNames.size() << " banks found." << std::endl;
    prog->setNumSteps(bankNames.size() + 5);

    //load event data
//    std::vector<int> totalEvents = getEventPixelId(file,true);
//    std::cout<<"total events are in banks: "<<totalEvents.size()<<std::endl;
//    std::cout<<"current path: "<<file->getPath()<<std::endl;

    std::vector<CSNSEventData> m_eventData = getEventData(file,true);
    std::cout<<"total events are in banks:"<<m_eventData.size()<<std::endl;



    //load instrument
    std::vector<std::vector<float>>pixelOffset = getPixelOffset(file,bankNames[0]);
    m_xOffset = pixelOffset[0];
    m_yOffset = pixelOffset[1];
    std::cout<<"xoffsets size is "<<m_xOffset.size()<<std::endl;
    std::vector<uint32_t> m_pixelId = getPixelId(file, bankNames[0]);
    std::cout<<"pixel_id size is "<<m_pixelId.size()<<std::endl;

    // Start with a dummy WS just to hold the logs and load the instrument
    prog->doReport("Creating workspace");
    // Bank workspace
    MatrixWorkspace_sptr WS1 = boost::dynamic_pointer_cast<MatrixWorkspace>( WorkspaceFactory::Instance().create("Workspace2D", m_bNumPixels, m_bNumBins + 1, m_bNumBins));
    // Monitor workspace
    MatrixWorkspace_sptr WS2 = boost::dynamic_pointer_cast<MatrixWorkspace>( WorkspaceFactory::Instance().create("Workspace2D", m_mNumPixels, m_mNumBins + 1, m_mNumBins));

    // Load log of source, chopper and SE
    prog->doReport("Loading time log");
    loadNXLog(file, WS1);
    //std::cout<<"current path: "<<file->getPath()<<std::endl;

    // close nxs file
    file->close();
    delete file;

    // Set to the output
    //setProperty("OutputWorkspace", boost::dynamic_pointer_cast<MatrixWorkspace>(WS));
    //setProperty("OutputWorkspace", WS);
    WorkspaceGroup_sptr wksp_group(new WorkspaceGroup);

    wksp_group->addWorkspace(WS1);
    wksp_group->addWorkspace(WS2);
    setProperty("OutputWorkspace",
                boost::dynamic_pointer_cast<Workspace>(wksp_group));

    delete prog;
}

} // namespace DataHandling
} // namespace Mantid
